@@ -6,21 +6,35 @@ import tornado.process
 
 from .routing import RoutingHandler
 
-from pyrox.tstream.iostream import IOStream, StreamClosedError
+from pyrox.tstream.iostream import IOHandler, StreamClosedError
 from pyrox.tstream.tcpserver import TCPServer
 
 from pyrox.log import get_logger
+from pyrox.about import VERSION
 from pyrox.http import (HttpRequest, HttpResponse, RequestParser,
                         ResponseParser, ParserDelegate)
+import traceback
 
 _LOG = get_logger(__name__)
 
-# Read 1k at a time
-_MAX_READ = 1024
+
+"""
+String representing a 0 length HTTP chunked encoding chunk.
+"""
 _CHUNK_CLOSE = b'0\r\n\r\n'
 
 
-def _write_to_stream(stream, data, is_chunked):
+"""
+Default return object on error. This should be configurable.
+"""
+_BAD_GATEWAY_RESP = HttpResponse()
+_BAD_GATEWAY_RESP.version = b'1.1'
+_BAD_GATEWAY_RESP.status = '502 Bad Gateway'
+_BAD_GATEWAY_RESP.header('Server').values.append('pyrox/{}'.format(VERSION))
+_BAD_GATEWAY_RESP.header('Content-Length').values.append('0')
+
+
+def _write_to_stream(stream, data, is_chunked, callback=None):
     if is_chunked:
         # Format and write this chunk
         chunk = bytearray()
@@ -28,10 +42,9 @@ def _write_to_stream(stream, data, is_chunked):
         chunk.extend('\r\n')
         chunk.extend(data)
         chunk.extend('\r\n')
-
-        stream.write(bytes(chunk))
+        stream.send(chunk, callback)
     else:
-        stream.write(data)
+        stream.send(data, callback)
 
 
 class AccumulationStream(object):
@@ -57,6 +70,7 @@ class ProxyHandler(ParserDelegate):
     def __init__(self, filter_pl, http_msg):
         self._filter_pl = filter_pl
         self._http_msg = http_msg
+        self._chunked = False
         self._last_header_field = None
         self._rejected = False
 
@@ -83,22 +97,13 @@ class DownstreamHandler(ProxyHandler):
         super(DownstreamHandler, self).__init__(filter_pl, HttpRequest())
         self._downstream = downstream
         self._upstream = None
-        self._body_fragment = None
+        self._preread_body = None
         self._connect_upstream = connect_upstream
 
-    def _store_body(self, body_fragment):
-        if self._body_fragment:
-            self._body_fragment.extend(body_fragment)
-        else:
-            self._body_fragment = body_fragment
-
-    def _get_body(self, next_body_fragment):
-        body_fragment = next_body_fragment
-        if self._body_fragment:
-            self._body_fragment.extend(body_fragment)
-            body_fragment = self._body_fragment
-            self._body_fragment = None
-        return body_fragment
+    def _store_chunk(self, body_fragment):
+        if not self._preread_body:
+            self._preread_body = bytearray()
+        self._preread_body.extend(body_fragment)
 
     def on_req_method(self, method):
         self._http_msg.method = method
@@ -110,18 +115,26 @@ class DownstreamHandler(ProxyHandler):
         # Execute against the pipeline
         action = self._filter_pl.on_request_head(self._http_msg)
 
-        # If there's a content length, negotiate the tansfer encoding
-        if self._http_msg.get_header('content-length'):
-            self._http_msg.remove_header('content-length')
-            self._http_msg.remove_header('transfer-encoding')
+        # If we are intercepting the request body do some negotiation
+        if self._filter_pl.intercepts_req_body():
+            self._chunked = True
 
-            self._http_msg.header('transfer-encoding').values.append('chunked')
+            # If there's a content length, negotiate the tansfer encoding
+            if self._http_msg.get_header('content-length'):
+                self._http_msg.remove_header('content-length')
+                self._http_msg.remove_header('transfer-encoding')
+
+                self._http_msg.header('transfer-encoding').values.append('chunked')
 
         # If we're rejecting then we're not going to connect to upstream
         if action.is_rejecting():
             self._rejected = True
             self._response = action.payload
         else:
+            # Hold up on the client side until we're done negotiating
+            # connections.
+            self._downstream.stop_recv()
+
             # We're routing to upstream; we need to know where to go
             if action.is_routing():
                 self._connect_upstream(self._http_msg, action.payload)
@@ -129,10 +142,16 @@ class DownstreamHandler(ProxyHandler):
                 self._connect_upstream(self._http_msg)
 
     def on_body(self, bytes, length, is_chunked):
+        self._chunked = is_chunked
+
+        if self._downstream.receiving():
+            # Hold up on the client side until we're done with this chunk
+            self._downstream.stop_recv()
+
         # Rejections simply discard the body
         if not self._rejected:
             accumulator = AccumulationStream()
-            data = bytes[:length]
+            data = bytes
 
             self._filter_pl.on_request_body(data, accumulator)
 
@@ -140,26 +159,32 @@ class DownstreamHandler(ProxyHandler):
                 data = accumulator.bytes
 
             if self._upstream:
-                # Write the content
-                _write_to_stream(
-                    self._upstream,
-                    self._get_body(data),
-                    is_chunked)
+                # When we write to the stream set the callback to resume
+                # reading from downstream.
+                _write_to_stream(self._upstream, data, is_chunked,
+                    self._downstream.resume_recv)
             else:
                 # If we're not connected upstream, store the fragment
                 # for later
-                self._store_body(data)
+                self._store_chunk(data)
+
+    def on_upstream_connect(self, upstream):
+        self._upstream = upstream
+
+        if self._preread_body and len(self._preread_body) > 0:
+            _write_to_stream(self._upstream, self._preread_body,
+                self._chunked, self._downstream.resume_recv)
+            self._preread_body = None
 
     def on_message_complete(self, is_chunked, keep_alive):
-        if self._rejected:
-            callback = None if keep_alive else self._downstream.close
+        if keep_alive:
+            self._http_msg = HttpRequest()
 
-            self._downstream.write(
-                self._response.to_bytes(),
-                callback=callback)
-        elif is_chunked:
-            print('closing downstream chunk')
-            self._upstream.write(_CHUNK_CLOSE)
+        if self._rejected:
+            self._downstream.send(self._response.to_bytes())
+        elif is_chunked or self._chunked:
+            # Finish the last chunk.
+            self._upstream.send(_CHUNK_CLOSE)
 
 
 class UpstreamHandler(ProxyHandler):
@@ -173,7 +198,6 @@ class UpstreamHandler(ProxyHandler):
         super(UpstreamHandler, self).__init__(filter_pl, HttpResponse())
         self._downstream = downstream
         self._upstream = upstream
-        self._chunked = False
 
     def on_status(self, status_code):
         self._http_msg.status = str(status_code)
@@ -181,58 +205,70 @@ class UpstreamHandler(ProxyHandler):
     def on_headers_complete(self):
         action = self._filter_pl.on_response_head(self._http_msg)
 
-        # If there's a content length, negotiate the tansfer encoding
-        if self._http_msg.get_header('content-length'):
-            self._chunked = True
-            self._http_msg.remove_header('content-length')
-            self._http_msg.remove_header('transfer-encoding')
+        # If we are intercepting the response body do some negotiation
+        if self._filter_pl.intercepts_resp_body():
 
-            self._http_msg.header('transfer-encoding').values.append('chunked')
+            # If there's a content length, negotiate the tansfer encoding
+            if self._http_msg.get_header('content-length'):
+                self._chunked = True
+                self._http_msg.remove_header('content-length')
+                self._http_msg.remove_header('transfer-encoding')
+
+                self._http_msg.header('transfer-encoding').values.append('chunked')
 
         if action.is_rejecting():
             self._rejected = True
             self._response = action.payload
         else:
-            self._downstream.write(self._http_msg.to_bytes())
+            self._downstream.send(self._http_msg.to_bytes())
 
     def on_body(self, bytes, length, is_chunked):
         # Rejections simply discard the body
         if not self._rejected:
             accumulator = AccumulationStream()
-            data = bytes[:length]
+            data = bytes
 
             self._filter_pl.on_response_body(data, accumulator)
 
             if accumulator.size() > 0:
                 data = accumulator.bytes
 
+            # Hold up on the upstream side until we're done sending this chunk
+            self._upstream.stop_recv()
+
+            # When we write to the stream set the callback to resume
+            # reading from upstream.
             _write_to_stream(
                 self._downstream,
                 data,
-                is_chunked or self._chunked)
+                is_chunked or self._chunked,
+                self._upstream.resume_recv)
 
     def on_message_complete(self, is_chunked, keep_alive):
-        callback = None if keep_alive else self._downstream.close
+        callback = self._upstream.close
+        self._upstream.stop_recv()
+
+        if keep_alive:
+            self._http_msg = HttpResponse()
+            callback = self._downstream.resume_recv
 
         if self._rejected:
-            self._downstream.write(
-                self._http_msg.to_bytes(),
-                callback=callback)
-
-        if is_chunked or self._chunked:
+            # Serialize our message to them
+            self._downstream.send(self._http_msg.to_bytes(), callback)
+        elif is_chunked or self._chunked:
             # Finish the last chunk.
-            self._downstream.write(
-                _CHUNK_CLOSE,
-                callback=callback)
-
+            self._downstream.send(_CHUNK_CLOSE, callback)
+        else:
+            callback()
 
 class ConnectionTracker(object):
 
-    def __init__(self, on_live_cb, pipe_broken_cb):
+    def __init__(self, on_stream_live, on_target_closed, on_target_error):
         self._streams = dict()
         self._target_in_use = None
-        self._on_live_cb = on_live_cb
-        self._pipe_broken_cb = pipe_broken_cb
+        self._on_stream_live = on_stream_live
+        self._on_target_closed = on_target_closed
+        self._on_target_error = on_target_error
 
     def destroy(self):
         for stream in self._streams.values():
@@ -245,28 +281,41 @@ class ConnectionTracker(object):
 
         if live_stream:
             # Make the cb ourselves since the socket's already connected
-            self._on_live_cb(live_stream)
+            self._on_stream_live(live_stream)
         else:
             self._new_connection(target)
 
     def _new_connection(self, target):
         # Set up our upstream socket
         us_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
-        us_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-        us_sock.setblocking(0)
 
-        # Create and bind the IOStream
-        live_stream = IOStream(us_sock)
+        # Create and bind the IOHandler
+        live_stream = IOHandler(us_sock)
         self._streams[target] = live_stream
 
         def on_close():
-            if self._target_in_use == target:
-                self._pipe_broken_cb()
+            # Disable error cb on close
+            live_stream.on_error(None)
+
             del self._streams[target]
-        live_stream.set_close_callback(on_close)
+            if self._target_in_use == target:
+                self.destroy()
+                self._on_target_closed()
+        live_stream.on_close(on_close)
+
+        def on_error(error):
+            # Dsiable close cb on error
+            live_stream.on_close(None)
+
+            if self._target_in_use == target:
+                del self._streams[target]
+                if self._target_in_use == target:
+                    self.destroy()
+                    self._on_target_error(error)
+        live_stream.on_error(on_error)
 
         def on_connect():
-            self._on_live_cb(live_stream)
+            self._on_stream_live(live_stream)
         live_stream.connect(target, on_connect)
 
 
@@ -282,7 +331,8 @@ class ProxyConnection(object):
         self._upstream_parser = None
         self._upstream_tracker = ConnectionTracker(
             self._on_upstream_live,
-            self._on_pipe_broken)
+            self._on_upstream_close,
+            self._on_upstream_error)
 
         # Setup all of the wiring for downstream
         self._downstream = downstream
@@ -291,10 +341,8 @@ class ProxyConnection(object):
             self._ds_filter_pl,
             self._connect_upstream)
         self._downstream_parser = RequestParser(self._downstream_handler)
-        self._downstream.set_close_callback(self._on_downstream_close)
-        self._downstream.read_until_close(
-            callback=self._on_downstream_read,
-            streaming_callback=self._on_downstream_read)
+        self._downstream.on_close(self._on_downstream_close)
+        self._downstream.on_recv(self._on_downstream_read)
 
     def _connect_upstream(self, request, route=None):
         if route:
@@ -309,7 +357,11 @@ class ProxyConnection(object):
         request.replace_header('host').values.append(
             '{}:{}'.format(upstream_target[0], upstream_target[1]))
         self._request = request
-        self._upstream_tracker.connect(upstream_target)
+
+        try:
+            self._upstream_tracker.connect(upstream_target)
+        except Exception as ex:
+            _LOG.exception(ex)
 
     def _on_upstream_live(self, upstream):
         self._upstream_handler = UpstreamHandler(
@@ -321,35 +373,39 @@ class ProxyConnection(object):
             self._upstream_parser.destroy()
         self._upstream_parser = ResponseParser(self._upstream_handler)
 
-        # Send the proxied request object
-        upstream.write(self._request.to_bytes())
+        # Set the read callback
+        upstream.on_recv(self._on_upstream_read)
 
-        # Set the read callback if it's not already reading
-        if not upstream.reading():
-            upstream.read_until_close(
-                callback=self._on_upstream_read,
-                streaming_callback=self._on_upstream_read)
+        # Send the proxied request object
+        upstream.send(self._request.to_bytes())
 
         # Drop the ref to the proxied request head
         self._request = None
 
         # Set up our downstream handler
-        self._downstream_handler._upstream = upstream
-
-        if not self._downstream.reading():
-            self._downstream.read_until_close(
-                callback=self._on_downstream_read,
-                streaming_callback=self._on_downstream_read)
+        self._downstream_handler.on_upstream_connect(upstream)
 
     def _on_downstream_close(self):
         self._upstream_tracker.destroy()
         self._downstream_parser.destroy()
+        self._downstream_parser = None
 
-    def _on_pipe_broken(self):
-        if not self._downstream.closed() and not self._downstream.closed():
+    def _on_downstream_error(self, error):
+        _LOG.error('Downstream error: {}'.format(error))
+        if not self._downstream.closed():
             self._downstream.close()
-        if self._upstream_parser:
+
+    def _on_upstream_error(self, error):
+        if not self._downstream.closed():
+            self._downstream.send(_BAD_GATEWAY_RESP.to_bytes())
+
+    def _on_upstream_close(self):
+        if not self._downstream.closed():
+            self._downstream.close()
+
+        if self._upstream_parser is not None:
             self._upstream_parser.destroy()
+            self._upstream_parser = None
 
     def _on_downstream_read(self, data):
         try:
